#include <stdio.h>
#include <sys/socket.h>
#include <pwd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <sys/syslog.h>

#include "config.h"
#include "libhttpd.h"

static httpd_server* hs = (httpd_server*) 0;
int terminate = 0;
static connecttab* connects;
int numconnects = 0, cnum;

static int maxconnects = MAXCLIENT;
char* user = SERVER_USER;
char* dir = DOCUMENT_ROOT;


static void handle_term( int sig )
{
	if ( hs != (httpd_server*) 0 )
		free_httpd_server( hs );

	for ( cnum = 0; cnum < maxconnects; cnum++ )
	{
		if ( connects[cnum].hc->file > 0 )
			close( connects[cnum].hc->file );
		free( connects[cnum].hc );
	}
	free( connects );
	exit( 0 );
}

int main( int argc, char** argv )
{
	httpd_sockaddr sa;
	fd_set active_fd_set, read_fd_set;
	int fd_max;
	int client_status, select_res;
	struct timeval tv;
	time_t now;

	struct passwd* pwd;
	uid_t uid;
	gid_t gid;

	int tv_sec = SELECT_TV_SEC;
	int tv_usec = SELECT_TV_USEC;

	sa.sa.sa_family = AF_INET;
	sa.sa_in.sin_addr.s_addr = htonl( INADDR_ANY );
	sa.sa_in.sin_port = htons( PORT );

	pwd = getpwnam( user );
	if ( pwd == (struct passwd*) 0 )
	{
		(void) fprintf( stderr, "unknown user - '%s'\n", user );
		exit( 1 );
	}
	uid = pwd->pw_uid;
	gid = pwd->pw_gid;

	switch ( fork() )
	{
		case 0: break;
		case -1:
			printf("fork() failed");
			exit( 1 );
		default: exit( 0 );
	}
	(void) setsid();
	(void) fclose( stdin );
	(void) fclose( stdout );
	(void) fclose( stderr );

	(void) signal( SIGTERM, handle_term );
	(void) signal( SIGINT, handle_term );
	(void) signal( SIGPIPE, SIG_IGN );
	(void) signal( SIGHUP, handle_term );
	(void) signal( SIGUSR1, handle_term );

	hs = httpd_initialize( &sa, PORT );
	if ( hs == (httpd_server*) 0 )
		exit( 1 );

	if ( getuid() == 0 )
	{
		if ( setgroups( 0, (const gid_t*) 0 ) < 0 )
		{
			syslog( LOG_CRIT, "setgroups - %m" );
			exit( 1 );
		}
		if ( setgid( gid ) < 0 )
		{
			syslog( LOG_CRIT, "setgid - %m" );
			exit( 1 );
		}
		if ( initgroups( user, gid ) < 0 )
			syslog( LOG_WARNING, "initgroups - %m" );
		if ( setuid( uid ) < 0 )
		{
			syslog( LOG_CRIT, "setuid - %m" );
			exit( 1 );
		}
	}

	connects = NEW( connecttab, maxconnects );
	if ( connects == (connecttab*) 0 )
	{
		syslog( LOG_CRIT, "out of memory" );
		exit( 1 );
	}

	for ( cnum = 0; cnum < maxconnects; cnum++ )
	{
		connects[cnum].conn_state = CNST_FREE;
		connects[cnum].hc = NEW( httpd_conn, 1 );
		conn_first_init( connects[cnum].hc );
	}

	FD_ZERO(&active_fd_set);
	FD_SET(hs->listen_fd, &active_fd_set);
	fd_max = hs->listen_fd+1;

	while ( ( ! terminate ) || numconnects > 0 )
	{
		time( &now );
		tv.tv_sec = tv_sec;
		tv.tv_usec = tv_usec;
		read_fd_set = active_fd_set;
		select_res = select ( fd_max, &read_fd_set, NULL, NULL, &tv );
		if ( select_res < 0 )
		{
			syslog( LOG_CRIT, "select error" );
			continue;
		}

		if ( numconnects > 0  )
		{
			fd_max = hs->listen_fd+1;
			for ( cnum = 0; cnum < maxconnects; cnum++ )
			{
				switch ( connects[cnum].conn_state )
				{
					case CNST_READING:
						if ( select_res > 0 && FD_ISSET( connects[cnum].hc->conn_fd, &read_fd_set ) )
						{
							client_status = httpd_got_request( connects[cnum].hc );
							if ( client_status < 0 )
							{
								FD_CLR(connects[cnum].hc->conn_fd, &active_fd_set);
								clear_client( connects, cnum, &numconnects );
								continue;
							}else if ( client_status == 1 ) {
								FD_CLR(connects[cnum].hc->conn_fd, &active_fd_set);
								connects[cnum].conn_state = CNST_SENDING;
								continue;
							}
						}
						if ( ( now - TIMEOUT ) >= connects[cnum].hc->now )
						{
							FD_CLR(connects[cnum].hc->conn_fd, &active_fd_set);
							clear_client( connects, cnum, &numconnects );
						}else if ( fd_max <= connects[cnum].hc->conn_fd ) fd_max = connects[cnum].hc->conn_fd+1;
					break;

					case CNST_SENDING :
						client_status = httpd_start_request( connects[cnum].hc );
						if ( client_status == -1 )
						{
							FD_CLR(connects[cnum].hc->conn_fd, &active_fd_set);
							clear_client( connects, cnum, &numconnects );
						}else if ( client_status == 1 ) {
							if ( connects[cnum].hc->keep_alive == 0 )
							{
								FD_CLR(connects[cnum].hc->conn_fd, &active_fd_set);
								clear_client( connects, cnum, &numconnects );
							}else{
								conn_init( connects[cnum].hc );
								connects[cnum].conn_state = CNST_READING;
								if ( fd_max <= connects[cnum].hc->conn_fd ) fd_max = connects[cnum].hc->conn_fd+1;
							}
						}
					break;
				}
			}
		}
		if ( FD_ISSET (hs->listen_fd, &read_fd_set) )
		{
			for ( cnum = 0; cnum < maxconnects; cnum++ )
				if ( connects[cnum].conn_state == CNST_FREE )
					break;
			if ( maxconnects  == cnum )
				continue;
			numconnects++;
			connects[cnum].conn_state = CNST_READING;
			if ( httpd_get_conn( hs->listen_fd, connects, cnum ) < 0 )
			{
				clear_client( connects, cnum, &numconnects );
				continue;
			}
			FD_SET( connects[cnum].hc->conn_fd, &active_fd_set);
			if ( fd_max <= connects[cnum].hc->conn_fd ) fd_max = connects[cnum].hc->conn_fd+1;
		}
	}
}
